@@ -24,12 +24,13 @@ from PIL import Image
 import keras
 import numpy as np
 import tensorflow as tf
-from keras.callbacks import EarlyStopping, TensorBoard, ModelCheckpoint
+from keras.callbacks import EarlyStopping, TensorBoard
 from keras.layers import Reshape, Conv2D, Input, Lambda, Dropout
 from keras.models import Model
 from keras.optimizers import Adam
 
 import cryolo.utils as utils
+import cryolo
 from cryolo.MultiGPUModelCheckpoint import MultiGPUModelCheckpoint
 from cryolo.ExtendedModelCheckpoint import ExtendedModelCheckpoint
 from cryolo.warmup_callback import WarmupCallback
@@ -38,14 +39,14 @@ from cryolo.backend import FullYoloFeature, CrYoloFeature, PhosaurusNetFeature
 from cryolo.patchwisebatchgenerator2 import PatchwiseBatchGenerator
 from cryolo.utils import BoundBox
 
-warmup_bs = keras.backend.variable(1.)
+warmup_bs = keras.backend.variable(1.0)
+
+
 class YOLO:
     def __init__(
         self,
         architecture,
         input_size,
-        input_depth,
-        labels,
         max_box_per_image,
         anchors,
         backend_weights=None,
@@ -54,6 +55,8 @@ class YOLO:
         pretrained_weights=None,
         fine_tune=False,
         num_fine_tune_layers=2,
+        input_depth=1,
+        labels=["particle"],
     ):
         self.input_size = input_size
         self.input_depth = input_depth
@@ -75,8 +78,12 @@ class YOLO:
 
         # make the feature extractor layers
         if not uniitest:
+            if type(self.input_size) is list:
+                self.input_size = self.input_size[::-1]
+            else:
+                self.input_size = [self.input_size,self.input_size]
             input_image = Input(
-                shape=(self.input_size, self.input_size, self.input_depth),
+                shape=(self.input_size[1], self.input_size[0], self.input_depth),
                 dtype="float32",
             )
             self.true_boxes = Input(shape=(1, 1, 1, max_box_per_image, 4))
@@ -106,7 +113,7 @@ class YOLO:
                     "Is your GPU currenlty used? Check it with nivdia-smi"
                 )
 
-            self.grid_h, self.grid_w = self.feature_extractor.get_output_shape()
+            self.grid_rows, self.grid_columns = self.feature_extractor.get_output_shape()
 
             features = self.feature_extractor.extract(input_image)
 
@@ -121,7 +128,7 @@ class YOLO:
                 kernel_initializer="lecun_normal",
             )(features)
             output = Reshape(
-                (self.grid_h, self.grid_w, self.nb_box, 4 + 1 + self.nb_class)
+                (self.grid_rows, self.grid_columns, self.nb_box, 4 + 1 + self.nb_class)
             )(output)
             output = Lambda(lambda args: args[0])([output, self.true_boxes])
             self.model = Model([input_image, self.true_boxes], output)
@@ -131,16 +138,16 @@ class YOLO:
             weights = layer.get_weights()
 
             new_kernel = np.random.normal(size=weights[0].shape) / (
-                self.grid_h * self.grid_w
+                    self.grid_rows * self.grid_columns
             )
             new_bias = np.random.normal(size=weights[1].shape) / (
-                self.grid_h * self.grid_w
+                    self.grid_rows * self.grid_columns
             )
 
             layer.set_weights([new_kernel, new_bias])
 
             if pretrained_weights is not None and os.path.exists(pretrained_weights):
-                print("Load weights: ", pretrained_weights)
+                print("\n Load weights: ", pretrained_weights)
                 self.load_weights(
                     weight_path=pretrained_weights,
                     do_fine_tune=self.fine_tune,
@@ -164,14 +171,16 @@ class YOLO:
 
         mask_shape = tf.shape(y_true)[:4]
 
-        # Generates a grid (e.g. 32 x 32)
+        # Generates a grid (e.g. 32 x 32) with coordinates
         cell_x = tf.to_float(
             tf.reshape(
-                tf.tile(tf.range(self.grid_w), [self.grid_h]),
-                (1, self.grid_h, self.grid_w, 1, 1),
+                tf.tile(tf.range(self.grid_columns), [self.grid_rows]),
+                (1, self.grid_rows, self.grid_columns, 1, 1),
             )
         )
-        cell_y = tf.transpose(cell_x, (0, 2, 1, 3, 4))
+        cell_y = tf.to_float(
+            tf.reshape(tf.tile(tf.range(self.grid_rows), [self.grid_columns]), (1, self.grid_columns, self.grid_rows, 1, 1)))
+        cell_y = tf.transpose(cell_y, (0, 2, 1, 3, 4))
 
         cell_grid = tf.tile(
             tf.concat([cell_x, cell_y], -1), [self.batch_size, 1, 1, self.nb_box, 1]
@@ -188,7 +197,7 @@ class YOLO:
         #
 
         # adjust x and y
-        pred_box_xy = tf.sigmoid(y_pred[..., :2]) + cell_grid
+        pred_box_xy = tf.sigmoid(y_pred[..., :2]) + cell_grid #coordinate corner + relative box position
 
         # adjust w and h
         pred_box_wh = tf.exp(y_pred[..., 2:4]) * np.reshape(
@@ -211,7 +220,7 @@ class YOLO:
         # adjust w and h
         true_box_wh = y_true[
             ..., 2:4
-        ]  # number of cells accross, horizontally and vertically
+        ]  # number of cells across, horizontally and vertically
 
         # adjust confidence
         true_wh_half = true_box_wh / 2.0
@@ -466,6 +475,7 @@ class YOLO:
             # If wanted, only train the last two layers of the network.
             ###############################
             if do_fine_tune:
+                print("Do fine tuning!")
                 for layer_index, layer in enumerate(self.model.layers):
                     if layer.name != "conv_23" and "model_" not in layer.name:
                         layer.trainable = False
@@ -522,7 +532,6 @@ class YOLO:
             free_layers = []
             for i in range(num_layers - 1):
                 layer_index = num_conv_layers - i
-                print("add conv_" + str(layer_index))
                 free_layers.append("conv_" + str(layer_index))
                 free_layers.append("conv_" + str(layer_index) + "_")
 
@@ -546,17 +555,19 @@ class YOLO:
         nms_threshold=0.3,
         num_patches=1,
         normalize_margin=0,
+        normalization="STANDARD"
     ):
 
         self.nms_threshold = nms_threshold
         self.obj_threshold = obj_threshold
         resized_normalized_array = None
-
+        norm_func = utils.get_normalization_func(normalization)
         for i in range(len(tiles_coord)):
             img_tmp = image_tiles[i, :, :]
+
             img_tmp = np.array(
                 Image.fromarray(img_tmp).resize(
-                    (self.input_size, self.input_size), resample=Image.BILINEAR
+                    (self.input_size[0], self.input_size[1]), resample=Image.BILINEAR
                 )
             )
             # cv2.resize(img_tmp, (self.input_size, self.input_size))
@@ -566,7 +577,7 @@ class YOLO:
                     shape=(image_tiles.shape[0], img_tmp.shape[0], img_tmp.shape[1])
                 )
 
-            resized_normalized_array[i, :, :] = self.feature_extractor.normalize(
+            resized_normalized_array[i, :, :] = norm_func(
                 img_tmp, margin_size=normalize_margin
             )
 
@@ -662,14 +673,10 @@ class YOLO:
                 ]
             )
 
-
-
-
             if (netout_index + 1) % nr_patches == 0:
                 boxes_per_image.append(boxes)
 
                 boxes = []
-
         boxes_nms = [
             non_maxima_suppress_fast(
                 boxes, self.nb_class, self.nms_threshold, min(0.1, obj_threshold)
@@ -791,7 +798,6 @@ class YOLO:
 
         return e_x / e_x.sum(axis, keepdims=True)
 
-
     def train(
         self,
         train_imgs,  # the list of images to train the model
@@ -806,6 +812,7 @@ class YOLO:
         no_object_scale,
         coord_scale,
         class_scale,
+        cryolo_mode,
         parallel_model=None,
         saved_weights_name="best_weights.h5",
         debug=False,
@@ -816,33 +823,45 @@ class YOLO:
         overlap_patches=0,
         num_cpus=-1,
         do_augmentation=True,
+        do_multiprocessing=True,
+        normalization="STANDARD"
     ):
 
         self.batch_size = batch_size
         self.warmup_epochs = warmup_epochs
-        global  warmup_bs
-        keras.backend.set_value(warmup_bs, warmup_epochs * (
-            train_times * (len(train_imgs) * num_patches * num_patches / batch_size + 1)
-            + valid_times
-            * (len(valid_imgs) * num_patches * num_patches / batch_size + 1)
-        ))
-        '''  
+        global warmup_bs
+        keras.backend.set_value(
+            warmup_bs,
+            warmup_epochs
+            * (
+                train_times
+                * (len(train_imgs) * num_patches * num_patches / batch_size + 1)
+                + valid_times
+                * (len(valid_imgs) * num_patches * num_patches / batch_size + 1)
+            ),
+        )
+        """  
         
-        '''
+        """
         self.object_scale = object_scale
         self.no_object_scale = no_object_scale
         self.coord_scale = coord_scale
         self.class_scale = class_scale
 
         self.debug = debug
+        self.do_multiprocessing = do_multiprocessing
 
-        #Create save weights path
+        # Create save weights path
+        saved_weights_filename, saved_weights_extension = os.path.splitext(
+            saved_weights_name
+        )
+        saved_weights_name = saved_weights_filename + "_tmp" + saved_weights_extension
         try:
             os.makedirs(os.path.dirname(saved_weights_name))
         except:
             pass
 
-        #if warmup_epochs > 0:
+        # if warmup_epochs > 0:
         #    nb_epoch = (
         #        warmup_epochs
         #    )  # if it's warmup stage, don't train more than warmup_epochs
@@ -867,10 +886,10 @@ class YOLO:
         ############################################
 
         generator_config = {
-            "IMAGE_H": self.input_size,
-            "IMAGE_W": self.input_size,
-            "GRID_H": self.grid_h,
-            "GRID_W": self.grid_w,
+            "IMAGE_H": self.input_size[1],
+            "IMAGE_W": self.input_size[0],
+            "GRID_H": self.grid_rows,
+            "GRID_W": self.grid_columns,
             "BOX": self.nb_box,
             "LABELS": self.labels,
             "CLASS": len(self.labels),
@@ -880,7 +899,7 @@ class YOLO:
             "NUM_PATCHES": num_patches,
         }
 
-        norm_func = utils.normalize
+        norm_func = utils.get_normalization_func(normalization)
 
         patch_assignements = np.ones(shape=(num_patches, num_patches))
 
@@ -888,6 +907,7 @@ class YOLO:
             train_imgs,
             generator_config,
             patch_assignements,
+            cryolo_mode=cryolo_mode,
             norm=norm_func,
             jitter=do_augmentation,
             overlap=overlap_patches,
@@ -899,6 +919,7 @@ class YOLO:
             valid_imgs,
             generator_config,
             patch_assignements,
+            cryolo_mode=cryolo_mode,
             norm=norm_func,
             jitter=False,
             overlap=overlap_patches,
@@ -912,8 +933,6 @@ class YOLO:
 
         all_callbacks = []
 
-
-
         early_stop = EarlyStopping(
             monitor="val_loss",
             min_delta=0.001,
@@ -928,27 +947,24 @@ class YOLO:
                 monitor="val_loss",
                 verbose=1,
                 save_best_only=True,
-                save_weights_only=False,
+                save_weights_only=True,
                 mode="min",
                 period=1,
             )
             checkpoint.set_template_model(self.model)
-            checkpoint.set_anchors(self.anchors)
-            checkpoint.set_fine_tune(self.fine_tune, self.num_fine_tune_layers)
         else:
             checkpoint = ExtendedModelCheckpoint(
                 saved_weights_name,
                 monitor="val_loss",
                 verbose=1,
                 save_best_only=True,
-                save_weights_only=False,
+                save_weights_only=True,
                 mode="min",
                 period=1,
             )
-            checkpoint.set_anchors(self.anchors)
-            checkpoint.set_fine_tune(self.fine_tune, self.num_fine_tune_layers)
-
-
+        checkpoint.set_anchors(self.anchors)
+        checkpoint.set_fine_tune(self.fine_tune, self.num_fine_tune_layers)
+        checkpoint.set_version(cryolo.__version__)
 
         reduceLROnPlateau = keras.callbacks.ReduceLROnPlateau(
             monitor="val_loss",
@@ -957,21 +973,19 @@ class YOLO:
             verbose=1,
         )
 
-
-
-        warmup_callback = WarmupCallback(warmup_bs=warmup_bs,
-                                         num_warmup_epochs=warmup_epochs,
-                                         len_train_img=len(train_imgs),
-                                         len_valid_img=len(valid_imgs),
-                                         batch_size=self.batch_size,
-                                         num_patches=num_patches,
-                                         train_times=train_times,
-                                         valid_times=valid_times,
-                                         early_stop_callback=early_stop,
-                                         modelcheckpoint=checkpoint,
-                                         reduceonnplateau=reduceLROnPlateau
-                                         )
-
+        warmup_callback = WarmupCallback(
+            warmup_bs=warmup_bs,
+            num_warmup_epochs=warmup_epochs,
+            len_train_img=len(train_imgs),
+            len_valid_img=len(valid_imgs),
+            batch_size=self.batch_size,
+            num_patches=num_patches,
+            train_times=train_times,
+            valid_times=valid_times,
+            early_stop_callback=early_stop,
+            modelcheckpoint=checkpoint,
+            reduceonnplateau=reduceLROnPlateau,
+        )
 
         ##APPEND CALLBACKS
 
@@ -1010,12 +1024,10 @@ class YOLO:
             all_callbacks.append(schedule)
 
         num_workers = multiprocessing.cpu_count() // 2
-        use_multiprocessing = True
         if num_cpus != -1:
             num_workers = num_cpus
-            if num_cpus == 1:
-                use_multiprocessing = False
-
+        #self.use_multiprocessing=False
+        num_workers = 1
         ############################################
         # Start the training process
         ############################################
@@ -1030,7 +1042,7 @@ class YOLO:
                 callbacks=all_callbacks,
                 workers=num_workers,
                 max_queue_size=multiprocessing.cpu_count(),
-                use_multiprocessing=use_multiprocessing,
+                use_multiprocessing=self.do_multiprocessing,
             )
         else:
             self.model.fit_generator(
@@ -1043,8 +1055,10 @@ class YOLO:
                 callbacks=all_callbacks,
                 workers=num_workers,
                 max_queue_size=multiprocessing.cpu_count(),
-                use_multiprocessing=use_multiprocessing,
+                use_multiprocessing=self.do_multiprocessing,
             )
+        # Rename to the user specified filename
+        os.rename(saved_weights_name, saved_weights_filename + saved_weights_extension)
 
 
 def non_maxima_suppress_fast(boxes, number_classes, nms_threshold, obj_threshold):
